"use strict";
/**
 * The source code present in this module was extracted from typescript-rest-swagger,
 * available at https://github.com/thiagobustamante/typescript-rest-swagger.git
 * under MIT licence, and modified to work as a stand alone typescript to swagger
 * type transformer. See the original licence below for more information.
 *
 * MIT License
 *
 * Copyright (c) Thiago da Rosa de Bustamante
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.getProperties = exports.getDecorators = exports.getDecoratorName = exports.getFirstMatchingJSDocTagName = exports.getLiteralValue = exports.getCommonPrimitiveAndArrayUnionType = exports.resolveType = exports.keywords = void 0;
const lodash_1 = __importDefault(require("lodash"));
const typescript_1 = __importDefault(require("typescript"));
const typescript_2 = require("typescript");
exports.keywords = [
    typescript_2.SyntaxKind.AnyKeyword,
    typescript_2.SyntaxKind.UnknownKeyword,
    typescript_2.SyntaxKind.NumberKeyword,
    typescript_2.SyntaxKind.BigIntKeyword,
    typescript_2.SyntaxKind.ObjectKeyword,
    typescript_2.SyntaxKind.BooleanKeyword,
    typescript_2.SyntaxKind.StringKeyword,
    typescript_2.SyntaxKind.SymbolKeyword,
    typescript_2.SyntaxKind.ThisKeyword,
    typescript_2.SyntaxKind.VoidKeyword,
    typescript_2.SyntaxKind.UndefinedKeyword,
    typescript_2.SyntaxKind.NullKeyword,
    typescript_2.SyntaxKind.NeverKeyword,
];
const syntaxKindMap = {};
syntaxKindMap[typescript_1.default.SyntaxKind.NumberKeyword] = 'number';
syntaxKindMap[typescript_1.default.SyntaxKind.StringKeyword] = 'string';
syntaxKindMap[typescript_1.default.SyntaxKind.BooleanKeyword] = 'boolean';
syntaxKindMap[typescript_1.default.SyntaxKind.VoidKeyword] = 'void';
const localReferenceTypeCache = {};
const inProgressTypes = {};
function resolveType(typeNode) {
    if (!typeNode) {
        return { typeName: 'void' };
    }
    const primitiveType = getPrimitiveType(typeNode);
    if (primitiveType) {
        return primitiveType;
    }
    if (typeNode.kind === typescript_1.default.SyntaxKind.ArrayType) {
        const arrayType = typeNode;
        return {
            elementType: resolveType(arrayType.elementType),
            typeName: 'array',
        };
    }
    if (typeNode.kind === typescript_1.default.SyntaxKind.AnyKeyword || typeNode.kind === typescript_1.default.SyntaxKind.ObjectKeyword) {
        return { typeName: 'object' };
    }
    if (typeNode.kind === typescript_1.default.SyntaxKind.TypeLiteral) {
        return getInlineObjectType(typeNode);
    }
    if (typeNode.kind === typescript_1.default.SyntaxKind.UnionType) {
        return getUnionType(typeNode);
    }
    if (typeNode.kind !== typescript_1.default.SyntaxKind.TypeReference) {
        throw new Error(`Unknown type: ${typescript_1.default.SyntaxKind[typeNode.kind]}`);
    }
    let typeReference = typeNode;
    let typeName = resolveSimpleTypeName(typeReference.typeName);
    if (typeName === 'Date') {
        return getDateType(typeNode);
    }
    if (typeName === 'Buffer') {
        return { typeName: 'buffer' };
    }
    if (typeName === 'DownloadBinaryData') {
        return { typeName: 'buffer' };
    }
    if (typeName === 'DownloadResource') {
        return { typeName: 'buffer' };
    }
    if (typeName === 'Promise') {
        typeReference = typeReference.typeArguments[0];
        return resolveType(typeReference);
    }
    if (typeName === 'Array') {
        typeReference = typeReference.typeArguments[0];
        return {
            elementType: resolveType(typeReference),
            typeName: 'array',
        };
    }
    let referenceType;
    if (typeReference.typeArguments && typeReference.typeArguments.length === 1) {
        const typeT = typeReference.typeArguments;
        referenceType = getReferenceType(typeReference.typeName, undefined, typeT);
    }
    else {
        referenceType = getReferenceType(typeReference.typeName);
    }
    return referenceType;
}
exports.resolveType = resolveType;
function getPrimitiveType(typeNode) {
    const primitiveType = syntaxKindMap[typeNode.kind];
    if (!primitiveType) {
        return undefined;
    }
    if (primitiveType === 'number') {
        const parentNode = typeNode.parent;
        if (!parentNode) {
            return { typeName: 'double' };
        }
        const validDecorators = ['IsInt', 'IsLong', 'IsFloat', 'IsDouble'];
        const jsdocTagName = getFirstMatchingJSDocTagName(parentNode, (tag) => {
            return validDecorators.some((t) => t === tag.tagName.text);
        });
        const decoratorName = getDecoratorName(parentNode, (identifier) => {
            return validDecorators.some((m) => m === identifier.text);
        });
        switch (decoratorName || jsdocTagName) {
            case 'IsInt':
                return { typeName: 'integer' };
            case 'IsLong':
                return { typeName: 'long' };
            case 'IsFloat':
                return { typeName: 'float' };
            case 'IsDouble':
                return { typeName: 'double' };
            default:
                return { typeName: 'double' };
        }
    }
    return { typeName: primitiveType };
}
function getDateType(typeNode) {
    const parentNode = typeNode.parent;
    if (!parentNode) {
        return { typeName: 'datetime' };
    }
    const decoratorName = getDecoratorName(parentNode, (identifier) => {
        return ['IsDate', 'IsDateTime'].some((m) => m === identifier.text);
    });
    switch (decoratorName) {
        case 'IsDate':
            return { typeName: 'date' };
        case 'IsDateTime':
            return { typeName: 'datetime' };
        default:
            return { typeName: 'datetime' };
    }
}
function getUnionType(typeNode) {
    const union = typeNode;
    let baseType = null;
    let isObject = false;
    union.types.forEach((type) => {
        if (baseType === null) {
            baseType = type;
        }
        if (baseType.kind !== type.kind) {
            isObject = true;
        }
    });
    if (isObject) {
        return { typeName: 'object' };
    }
    return {
        enumMembers: union.types.map((type, index) => {
            return type.getText() ? removeQuotes(type.getText()) : index;
        }),
        typeName: 'enum',
    };
}
function removeQuotes(str) {
    return str.replace(/^["']|["']$/g, '');
}
function getInlineObjectType(typeNode) {
    const type = {
        properties: getModelTypeProperties(typeNode),
        typeName: '',
    };
    return type;
}
function getReferenceType(type, genericTypeMap, genericTypes) {
    let typeName = resolveFqTypeName(type);
    if (genericTypeMap && genericTypeMap.has(typeName)) {
        const refType = genericTypeMap.get(typeName);
        type = refType.typeName;
        typeName = resolveFqTypeName(type);
    }
    const typeNameWithGenerics = getTypeName(typeName, genericTypes);
    try {
        const existingType = localReferenceTypeCache[typeNameWithGenerics];
        if (existingType) {
            return existingType;
        }
        inProgressTypes[typeNameWithGenerics] = true;
        const modelTypeDeclaration = getModelTypeDeclaration(type);
        const properties = getModelTypeProperties(modelTypeDeclaration, genericTypes);
        const additionalProperties = getModelTypeAdditionalProperties(modelTypeDeclaration);
        const referenceType = {
            description: '',
            properties: properties,
            typeName: typeNameWithGenerics,
        };
        if (additionalProperties && additionalProperties.length) {
            referenceType.additionalProperties = additionalProperties;
        }
        localReferenceTypeCache[typeNameWithGenerics] = referenceType;
        return referenceType;
    }
    catch (err) {
        console.error(`There was a problem resolving type of '${getTypeName(typeName, genericTypes)}'.`);
        throw err;
    }
}
function resolveFqTypeName(type) {
    if (type.kind === typescript_1.default.SyntaxKind.Identifier) {
        return type.text;
    }
    const qualifiedType = type;
    return resolveFqTypeName(qualifiedType.left) + '.' + qualifiedType.right.text;
}
function resolveSimpleTypeName(type) {
    if (type.kind === typescript_1.default.SyntaxKind.Identifier) {
        return type.text;
    }
    const qualifiedType = type;
    return qualifiedType.right.text;
}
function getTypeName(typeName, genericTypes) {
    if (!genericTypes || !genericTypes.length) {
        return typeName;
    }
    return typeName + genericTypes.map((t) => getAnyTypeName(t)).join('');
}
function getAnyTypeName(typeNode) {
    const primitiveType = syntaxKindMap[typeNode.kind];
    if (primitiveType) {
        return primitiveType;
    }
    if (typeNode.kind === typescript_1.default.SyntaxKind.ArrayType) {
        const arrayType = typeNode;
        return getAnyTypeName(arrayType.elementType) + 'Array';
    }
    if (typeNode.kind === typescript_1.default.SyntaxKind.UnionType || typeNode.kind === typescript_1.default.SyntaxKind.AnyKeyword) {
        return 'object';
    }
    if (typeNode.kind !== typescript_1.default.SyntaxKind.TypeReference) {
        throw new Error(`Unknown type: ${typescript_1.default.SyntaxKind[typeNode.kind]}`);
    }
    const typeReference = typeNode;
    try {
        const typeName = typeReference.typeName.text;
        if (typeName === 'Array') {
            return getAnyTypeName(typeReference.typeArguments[0]) + 'Array';
        }
        return typeName;
    }
    catch (e) {
        console.error(e);
        return typeNode.toString();
    }
}
function nodeIsUsable(node) {
    switch (node.kind) {
        case typescript_1.default.SyntaxKind.InterfaceDeclaration:
        case typescript_1.default.SyntaxKind.ClassDeclaration:
        case typescript_1.default.SyntaxKind.TypeAliasDeclaration:
            return true;
        default:
            return false;
    }
}
function getModelTypeDeclaration(type) {
    // @ts-ignore
    const statements = global.allNodes;
    const typeName = type.kind === typescript_1.default.SyntaxKind.Identifier ? type.text : type.right.text;
    const modelTypes = statements.filter((node) => {
        if (!nodeIsUsable(node)) {
            return false;
        }
        const modelTypeDeclaration = node;
        return modelTypeDeclaration.name.text === typeName;
    });
    if (!modelTypes.length) {
        throw new Error(`No matching model found for referenced type ${typeName}`);
    }
    return modelTypes[0];
}
function getModelTypeProperties(node, genericTypes) {
    if (node.kind === typescript_1.default.SyntaxKind.TypeLiteral || node.kind === typescript_1.default.SyntaxKind.InterfaceDeclaration) {
        const interfaceDeclaration = node;
        return interfaceDeclaration.members
            .filter((member) => {
            if (member.type && member.type.kind === typescript_1.default.SyntaxKind.FunctionType) {
                return false;
            }
            return member.kind === typescript_1.default.SyntaxKind.PropertySignature;
        })
            .map((member) => {
            const propertyDeclaration = member;
            const identifier = propertyDeclaration.name;
            if (!propertyDeclaration.type) {
                throw new Error('No valid type found for property declaration.');
            }
            // Declare a variable that can be overridden if needed
            let aType = propertyDeclaration.type;
            // aType.kind will always be a TypeReference when the property of Interface<T> is of type T
            if (aType.kind === typescript_1.default.SyntaxKind.TypeReference && genericTypes && genericTypes.length && node.typeParameters) {
                // The type definitions are conviently located on the object which allow us to map -> to the genericTypes
                const typeParams = lodash_1.default.map(node.typeParameters, (typeParam) => {
                    return typeParam.name.text;
                });
                // I am not sure in what cases
                const typeIdentifier = aType.typeName;
                let typeIdentifierName;
                // typeIdentifier can either be a Identifier or a QualifiedName
                if (typeIdentifier.text) {
                    typeIdentifierName = typeIdentifier.text;
                }
                else {
                    typeIdentifierName = typeIdentifier.right.text;
                }
                // I could not produce a situation where this did not find it so its possible this check is irrelevant
                const indexOfType = lodash_1.default.indexOf(typeParams, typeIdentifierName);
                if (indexOfType >= 0) {
                    aType = genericTypes[indexOfType];
                }
            }
            return {
                description: '',
                name: identifier.text,
                required: !propertyDeclaration.questionToken,
                type: resolveType(aType),
            };
        });
    }
    if (node.kind === typescript_1.default.SyntaxKind.TypeAliasDeclaration) {
        const typeAlias = node;
        return !exports.keywords.includes(typeAlias.type.kind) ? getModelTypeProperties(typeAlias.type, genericTypes) : [];
    }
    const classDeclaration = node;
    let properties = classDeclaration.members.filter((member) => {
        if (member.kind !== typescript_1.default.SyntaxKind.PropertyDeclaration) {
            return false;
        }
        const propertySignature = member;
        return propertySignature && hasPublicMemberModifier(propertySignature);
    });
    const classConstructor = classDeclaration.members.find((member) => member.kind === typescript_1.default.SyntaxKind.Constructor);
    if (classConstructor && classConstructor.parameters) {
        properties = properties.concat(classConstructor.parameters.filter((parameter) => hasPublicConstructorModifier(parameter)));
    }
    return properties.map((declaration) => {
        const identifier = declaration.name;
        if (!declaration.type) {
            throw new Error('No valid type found for property declaration.');
        }
        return {
            description: '',
            name: identifier.text,
            required: !declaration.questionToken,
            type: resolveType(resolveTypeParameter(declaration.type, classDeclaration, genericTypes)),
        };
    });
}
function resolveTypeParameter(type, classDeclaration, genericTypes) {
    if (genericTypes && classDeclaration.typeParameters && classDeclaration.typeParameters.length) {
        for (let i = 0; i < classDeclaration.typeParameters.length; i++) {
            if (type.typeName && classDeclaration.typeParameters[i].name.text === type.typeName.text) {
                return genericTypes[i];
            }
        }
    }
    return type;
}
function getModelTypeAdditionalProperties(node) {
    if (node.kind === typescript_1.default.SyntaxKind.InterfaceDeclaration) {
        const interfaceDeclaration = node;
        return interfaceDeclaration.members
            .filter((member) => member.kind === typescript_1.default.SyntaxKind.IndexSignature)
            .map((member) => {
            const indexSignatureDeclaration = member;
            const indexType = resolveType(indexSignatureDeclaration.parameters[0].type);
            if (indexType.typeName !== 'string') {
                throw new Error(`Only string indexers are supported. Found ${indexType.typeName}.`);
            }
            return {
                description: '',
                name: '',
                required: true,
                type: resolveType(indexSignatureDeclaration.type),
            };
        });
    }
    return undefined;
}
function hasPublicMemberModifier(node) {
    return (!node.modifiers ||
        node.modifiers.every((modifier) => {
            return modifier.kind !== typescript_1.default.SyntaxKind.ProtectedKeyword && modifier.kind !== typescript_1.default.SyntaxKind.PrivateKeyword;
        }));
}
function hasPublicConstructorModifier(node) {
    return (node.modifiers &&
        node.modifiers.some((modifier) => {
            return modifier.kind === typescript_1.default.SyntaxKind.PublicKeyword;
        }));
}
function getCommonPrimitiveAndArrayUnionType(typeNode) {
    if (typeNode && typeNode.kind === typescript_1.default.SyntaxKind.UnionType) {
        const union = typeNode;
        const types = union.types.map((t) => resolveType(t));
        const arrType = types.find((t) => t.typeName === 'array');
        const primitiveType = types.find((t) => t.typeName !== 'array');
        if (types.length === 2 &&
            arrType &&
            arrType.elementType &&
            primitiveType &&
            arrType.elementType.typeName === primitiveType.typeName) {
            return arrType;
        }
    }
    return null;
}
exports.getCommonPrimitiveAndArrayUnionType = getCommonPrimitiveAndArrayUnionType;
function getLiteralValue(expression) {
    if (expression.kind === typescript_1.default.SyntaxKind.StringLiteral) {
        return expression.text;
    }
    if (expression.kind === typescript_1.default.SyntaxKind.NumericLiteral) {
        return parseFloat(expression.text);
    }
    if (expression.kind === typescript_1.default.SyntaxKind.TrueKeyword) {
        return true;
    }
    if (expression.kind === typescript_1.default.SyntaxKind.FalseKeyword) {
        return false;
    }
    if (expression.kind === typescript_1.default.SyntaxKind.ArrayLiteralExpression) {
        return expression.elements.map((e) => getLiteralValue(e));
    }
    return;
}
exports.getLiteralValue = getLiteralValue;
function getFirstMatchingJSDocTagName(node, isMatching) {
    const tags = getMatchingJSDocTags(node, isMatching);
    if (!tags || !tags.length) {
        return undefined;
    }
    return tags[0].tagName.text;
}
exports.getFirstMatchingJSDocTagName = getFirstMatchingJSDocTagName;
function getMatchingJSDocTags(node, isMatching) {
    const jsDocs = node.jsDoc;
    if (!jsDocs || !jsDocs.length) {
        return undefined;
    }
    const jsDoc = jsDocs[0];
    if (!jsDoc.tags) {
        return undefined;
    }
    return jsDoc.tags.filter(isMatching);
}
function getDecoratorName(node, isMatching) {
    const decorator = getDecorator(node, isMatching);
    return decorator ? decorator.text : undefined;
}
exports.getDecoratorName = getDecoratorName;
function getDecorator(node, isMatching) {
    const decorators = getDecorators(node, isMatching);
    if (!decorators || !decorators.length) {
        return undefined;
    }
    return decorators[0];
}
function getDecorators(node, isMatching) {
    const decorators = node.decorators;
    if (!decorators || !decorators.length) {
        return [];
    }
    return decorators
        .map((d) => {
        const result = {
            arguments: [],
            typeArguments: [],
        };
        let x = d.expression;
        if (typescript_1.default.isCallExpression(x)) {
            if (x.arguments) {
                result.arguments = x.arguments.map((argument) => {
                    if (typescript_1.default.isStringLiteral(argument)) {
                        return argument.text;
                    }
                    else if (typescript_1.default.isNumericLiteral(argument)) {
                        return argument.text;
                    }
                    else {
                        return argument;
                    }
                });
            }
            if (x.typeArguments) {
                result.typeArguments = x.typeArguments;
            }
            x = x.expression;
        }
        result.text = x.text || x.name.text;
        return result;
    })
        .filter(isMatching);
}
exports.getDecorators = getDecorators;
function getProperties(properties) {
    const swaggerProperties = {};
    properties.forEach((property) => {
        const swaggerType = getSwaggerType(property.type);
        if (!swaggerType.$ref) {
            swaggerType.description = property.description;
        }
        swaggerProperties[property.name] = swaggerType;
    });
    return swaggerProperties;
}
exports.getProperties = getProperties;
function getSwaggerType(type) {
    const swaggerType = getSwaggerTypeForPrimitiveType(type);
    if (swaggerType) {
        return swaggerType;
    }
    const arrayType = type;
    if (arrayType.elementType) {
        return getSwaggerTypeForArrayType(arrayType);
    }
    const enumType = type;
    if (enumType.enumMembers) {
        return getSwaggerTypeForEnumType(enumType);
    }
    const refType = type;
    if (refType.properties && refType.description !== undefined) {
        return getSwaggerTypeForReferenceType(type);
    }
    const objectType = type;
    return getSwaggerTypeForObjectType(objectType);
}
function getSwaggerTypeForPrimitiveType(type) {
    const typeMap = {
        binary: { type: 'string', format: 'binary' },
        boolean: { type: 'boolean' },
        buffer: { type: 'file' },
        byte: { type: 'string', format: 'byte' },
        date: { type: 'string', format: 'date' },
        datetime: { type: 'string', format: 'date-time' },
        double: { type: 'number', format: 'double' },
        file: { type: 'file' },
        float: { type: 'number', format: 'float' },
        integer: { type: 'integer', format: 'int32' },
        long: { type: 'integer', format: 'int64' },
        object: { type: 'object' },
        string: { type: 'string' },
        void: { type: 'void' },
    };
    return typeMap[type.typeName];
}
function getSwaggerTypeForObjectType(objectType) {
    return { type: 'object', properties: getProperties(objectType.properties) };
}
function getSwaggerTypeForArrayType(arrayType) {
    return { type: 'array', items: getSwaggerType(arrayType.elementType) };
}
function getSwaggerTypeForEnumType(enumType) {
    function getDerivedTypeFromValues(values) {
        return values.reduce((derivedType, item) => {
            const currentType = typeof item;
            derivedType = derivedType && derivedType !== currentType ? 'string' : currentType;
            return derivedType;
        }, null);
    }
    const enumValues = enumType.enumMembers.map((member) => member);
    return {
        enum: enumType.enumMembers.map((member) => member),
        type: getDerivedTypeFromValues(enumValues),
    };
}
function getSwaggerTypeForReferenceType(referenceType) {
    return { $ref: `#/definitions/${referenceType.typeName}` };
}
