/// <reference types="koa__cors" />
import cors from '@koa/cors';
import { APIGatewayProxyEvent, APIGatewayProxyResult, Context } from 'aws-lambda';
export interface Router {
    addRoutes: (routes: Route[]) => void;
    route: (request: APIGatewayProxyEvent, context: Context) => Promise<APIGatewayProxyResult>;
}
export interface Route {
    path: string;
    get?: RouteConfig;
    post?: RouteConfig;
    put?: RouteConfig;
    patch?: RouteConfig;
    delete?: RouteConfig;
}
export interface RouteConfig {
    handler: RouteHandler;
    matchedPath: string;
    matchedMethod: string;
    tracingConfig: {
        logRequest: boolean;
        logResponse: boolean;
    };
}
export declare type RouteHandler = (request: HttpRequest<any, any, any, any>) => Promise<HttpResponse<unknown>>;
export interface HttpRequest<PathSchema extends BasePathSchema = BasePathSchema, BodySchema extends BaseBodySchema = BaseBodySchema, QuerySchema extends BaseQuerySchema = BaseQuerySchema, HeadersSchema extends BaseHeadersSchema = BaseHeadersSchema, MultiValueQuerySchema extends BaseMultiValueQuerySchema = BaseMultiValueQuerySchema> {
    path: PathSchema;
    body: BodySchema;
    query: QuerySchema;
    headers: HeadersSchema;
    multiValueQuery: MultiValueQuerySchema;
    queryString: string;
}
export declare type MethodRequest<PathSchema extends BasePathSchema = BasePathSchema, BodySchema extends BaseBodySchema = BaseBodySchema, QuerySchema extends BaseQuerySchema = BaseQuerySchema, HeadersSchema extends BaseHeadersSchema = BaseHeadersSchema, MultiValueQuerySchema extends BaseMultiValueQuerySchema = BaseMultiValueQuerySchema> = HttpRequest<PathSchema, BodySchema, QuerySchema, HeadersSchema, MultiValueQuerySchema>;
export interface HttpResponse<T> {
    statusCode: number;
    body?: T;
    headers?: Record<string, string>;
}
export declare type MethodResponse<T> = {
    statusCode: number;
    body?: T;
    headers?: Record<string, string>;
} | T;
export declare type BasePathSchema = Record<string, string | undefined>;
export declare type BaseBodySchema = Record<string, unknown> | string | undefined;
export declare type BaseQuerySchema = Record<string, string | undefined>;
export declare type BaseMultiValueQuerySchema = Record<string, string[] | undefined>;
export declare type BaseHeadersSchema = Record<string, string | undefined> & {
    getHeader: (header: string) => string | undefined;
};
export interface CorsConfig {
    enabled: boolean;
    config: cors.Options;
}
